# -*- coding: utf-8 -*-
"""
@author: Rafael
"""
import io
import warnings
from typing import Dict, List, Tuple, Union, Optional

import pandas as pd
import polars as pl
import requests
import yfinance as yf
import tesouro_direto_br as tesouro_direto

warnings.filterwarnings("ignore")

pd.set_option("display.float_format", lambda x: "%.6f" % x)
pd.set_option("display.max_rows", 100)
pd.set_option("display.max_columns", 10)
pd.set_option("display.width", 1000)

import seaborn as sns; sns.set()
import matplotlib.pyplot as plt

def get_cdi(inicio: str, fim: str, metodo_cdi: str ='bacen',  proxies: Optional[Dict[str, str]] = None) -> pd.DataFrame:
    '''Função que provê o retorno do CDI usada como referência especialmente para Renda Fixa.
    Esta função usa os métodos: tesouro, anbima ou bacen para extrair a rentabilidade.
    Tesouro = usa-se o rendimento diário dos titulos Tesouro Selic de vencimentos mais longos;
    Anbima = usa-se o índice IMA-S que usa rendimentos de Tesouro Selic;
    BACEN = usa-se a série diária do CDI.
    As datas devem ser no formato string '2025-01-02', ou seja, 'ANO-MES-DIA'.
    A saída gera o retorno diário e o acumulado.
    Os valores resultantes são próximos, mas não iguais. Divergem em função da metodologia.'''
    if metodo_cdi.lower()=='tesouro':
        titulos_ofertados = tesouro_direto.busca_tesouro_direto(tipo="taxa", proxies=proxies, agrupar=True).reset_index()
    
        excluir = ["Juros Semestrais", "Renda+", "Educa+"]
        titulos_ofertados_filtrado = titulos_ofertados[(titulos_ofertados["Data Base"]>=inicio) &
                                           (titulos_ofertados["Data Base"]<fim) &
                                           (~titulos_ofertados["Tipo Titulo"].str.contains("&".join(excluir)))].set_index(["Tipo Titulo",
                                                                                            "Data Vencimento"])
        selic = titulos_ofertados_filtrado.loc[["Tesouro Selic"], :].sort_values("Data Base", ascending=False)
        selic_longa = selic.reset_index().groupby(["Data Base"])[["Data Vencimento", 
                                                    'Taxa Compra Manha', 'Taxa Venda Manha', 
                                                    'PU Compra Manha', 'PU Venda Manha', 'PU Base Manha']].max()
        cdi = selic_longa[["PU Base Manha"]]
        cdi.columns = ['CDI']
        cdi['Retorno CDI'] = cdi["CDI"].pct_change()
        cdi['Retorno Acumulado CDI'] = (1+cdi['Retorno CDI']).cumprod()-1
    elif metodo_cdi.lower()=='anbima':
        imas = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IMAS-HISTORICO.xls',
                             engine='calamine', columns=[1,2])
        imas = imas.filter((pl.col('Data de Referência')>=pd.to_datetime(inicio)) & (pl.col('Data de Referência')<=pd.to_datetime(fim)))
        imas = imas.rename({'Número Índice': 'CDI'})
        imas = imas.with_columns(pl.col('CDI').pct_change().alias("Retorno CDI"))
        imas = imas.with_columns(((pl.col('Retorno CDI')+1).cum_prod()-1).alias("Retorno Acumulado CDI"))
        cdi = imas.to_pandas().set_index('Data de Referência')
    elif metodo_cdi.lower()=="bacen":
        codigo_bcb = 12
        anoi, mesi, diai = inicio.split('-')
        anof, mesf, diaf = fim.split('-')
        iniciob = '/'.join([diai, mesi, anoi])
        fimb = '/'.join([diaf, mesf, anof])
        url = f"https://api.bcb.gov.br/dados/serie/bcdata.sgs.{codigo_bcb}/dados?formato=json&dataInicial={iniciob}&dataFinal={fimb}"
        if proxies:
            dados_bacen = requests.get(url, proxies=proxies, verify=False).text
        else:
            dados_bacen = requests.get(url).text
        cdi = pd.read_json(io.StringIO(dados_bacen))
        cdi["data"] = pd.to_datetime(cdi["data"], dayfirst=True)
        cdi = cdi.set_index("data")
        cdi = cdi[(cdi.index>=inicio) & (cdi.index<=fim)] / 100
        cdi.columns = ['CDI']
        cdi['Retorno CDI'] = cdi["CDI"]
        cdi['Retorno Acumulado CDI'] = (1+cdi['Retorno CDI']).cumprod()-1
    else:
        raise ValueError("Método não permitido")
    return cdi

def get_indices_anbima(data_inicio: str, data_fim: str, benchmark: str = "imas") -> pd.DataFrame:
    '''Função que provê o retorno de alguns indices ANBIMA usada como referência, especialmente para Renda Fixa.
    Esta função implementa os seguintes índices: 
    -IMA-S (imas);
    -IMA-B5 (imab5);
    -IMA-B5+ (imab5+)
    -IMA-B5 P2 (imab5p2);
    -IRFM (irfm);
    -IRFM P2 (irfmp2);
    -IHFA (ihfa).
    As datas devem ser no formato string '2025-01-02', ou seja, 'ANO-MES-DIA'.
    Mais informações em https://data.anbima.com.br/indices'''
    if benchmark.lower()=="imas":
        indice = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IMAS-HISTORICO.xls',
                             engine='calamine', columns=[1,2])
    elif benchmark.lower()=="imab5":
        indice = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IMAB5-HISTORICO.xls',
                     engine='calamine', columns=[1,2])
    elif benchmark.lower()=="imab5+":
        indice = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IMAB5MAIS-HISTORICO.xls',
                     engine='calamine', columns=[1,2])
    elif benchmark.lower()=="imab5p2":
        indice = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IMAB5P2-HISTORICO.xls',
                     engine='calamine', columns=[1,2])
    elif benchmark.lower()=="irfm":
        indice = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IRFM-HISTORICO.xls',
                     engine='calamine', columns=[1,2])
    elif benchmark.lower()=="irfmp2":
        indice = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IRFMP2-HISTORICO.xls',
                     engine='calamine', columns=[1,2])
    elif benchmark.lower()=="ihfa":
        indice = pl.read_excel('https://s3-data-prd-use1-precos.s3.us-east-1.amazonaws.com/arquivos/indices-historico/IHFA-HISTORICO.xls',
                     engine='calamine', columns=[1,2])
    else:
        raise ValueError('Benchmark não encontrado.')
    indice = indice.filter((pl.col('Data de Referência')>=pd.to_datetime(data_inicio)) & (pl.col('Data de Referência')<=pd.to_datetime(data_fim)))
    indice = indice.rename({'Número Índice': benchmark.upper()})
    return indice.to_pandas().set_index('Data de Referência')

def get_benchmarks(data_inicio: str, data_fim: str, benchmark: str = "CDI", metodo_cdi = 'bacen', proxy=None) -> pd.DataFrame:
    '''Função que provê o retorno de alguns indices ANBIMA, CDI e renda variável: ibov, divo11 (similar ao IDIV) e sp500.
    Esta função implementa os seguintes índices anbima: 
    -IMA-S (imas);
    -IMA-B5 (imab5);
    -IMA-B5+ (imab5+)
    -IMA-B5 P2 (imab5p2);
    -IRFM (irfm);
    -IRFM P2 (irfmp2);
    -IHFA (ihfa).
    As datas devem ser no formato string '2025-01-02', ou seja, 'ANO-MES-DIA'.
    A saída gera o retorno diário e o acumulado.'''
    if proxy: yf.set_config(proxy=proxy)
    if benchmark.upper()=="CDI":
        df_benchmark = get_cdi(data_inicio, data_fim, metodo_cdi, proxy)
    else:
        if benchmark.upper()=="IBOV":
            df_benchmark = yf.download("^BVSP", start=data_inicio, end=data_fim, interval="1d", auto_adjust=True)["Close"]
            df_benchmark.columns = ["IBOV"]
        elif benchmark.upper()=="DIVO11":
            df_benchmark = yf.download("DIVO11.SA", start=data_inicio, end=data_fim, interval="1d", auto_adjust=True)["Close"]
            df_benchmark.columns = ["DIVO11"]
        elif benchmark.upper()=="SP500":
            df_benchmark = yf.download("^GSPC", start=data_inicio, end=data_fim, interval="1d", auto_adjust=True)["Close"]
            df_benchmark.columns = ["SP500"]
        elif benchmark.upper()=="USD":
            df_benchmark = yf.download("BRL=X", start=data_inicio, end=data_fim, interval="1d", auto_adjust=True)["Close"]
            df_benchmark.columns = ["USD"]
        else:
            df_benchmark = get_indices_anbima(data_inicio, data_fim, benchmark)
        df_benchmark[f'Retorno {benchmark.upper()}'] = df_benchmark[benchmark.upper()].pct_change()
        df_benchmark[f'Retorno Acumulado {benchmark.upper()}'] = (1+df_benchmark[f'Retorno {benchmark.upper()}']).cumprod()-1
    return df_benchmark

def get_stocks(
                acoes: Union[List[str], str],
                data_inicio: str,
                data_fim: str,
                proxy: Union[Dict[str, str], None] = None,
                ) -> pd.DataFrame:
    """Função para capturar dados de Ações ou Índices Listados.
    """
    if proxy: yf.set_config(proxy=proxy)
    df1 = pd.DataFrame()
    if isinstance(acoes, list):
        for st in acoes:
            if not st.endswith(".SA"): st = st+".SA"
            if proxy:
                df = yf.download(st, start=data_inicio, end=data_fim, auto_adjust=True)['Close']
            else:
                df = yf.download(st, start=data_inicio, end=data_fim)['Close']
            df1 = pd.concat([df, df1], axis=1)
    else:
        if not acoes.endswith(".SA"): acoes = acoes+".SA"
        if proxy:
            df1 = yf.download(acoes, start=data_inicio, end=data_fim, auto_adjust=True)['Close']
        else:
            df1 = yf.download(acoes, start=data_inicio, end=data_fim)['Close']
    df1.index = pd.to_datetime(df1.index)
    for cols in df1.columns:
        df1[f'Retorno {cols}'] = df1[cols].pct_change()
        df1[f'Retorno Acumulado {cols}'] = (1+df1[f'Retorno {cols}']).cumprod()-1
    return df1


# -*- coding: utf-8 -*-
"""
@author: Rafael
"""
from comparar_fundos_br.benchmarks import *
import warnings
from typing import Any, List, Tuple, Union, Optional
from tqdm import tqdm
import numpy as np
import pandas as pd
from itertools import cycle

warnings.filterwarnings("ignore")

pd.set_option("display.float_format", lambda x: "%.6f" % x)
pd.set_option("display.max_rows", 100)
pd.set_option("display.max_columns", 10)
pd.set_option("display.width", 1000)

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, rgb_to_hsv

def _get_valores_iniciais(df: pd.DataFrame) -> List[float]:
    if isinstance(df, pd.DataFrame):
        return [df[df[col].notnull()][col].iloc[0] for col in df.columns]
    else:
        return [value for value in df.values if np.isnan(value)==False][0]
    
def _get_valores_finais(df: pd.DataFrame) -> List[float]:
    if isinstance(df, pd.DataFrame):
        return [df[df[col].notnull()][col].iloc[-1] for col in df.columns]
    else:
        return [value for value in df.values if np.isnan(value)==False][-1]

def _get_cotas_normalizadas(df: pd.DataFrame) -> pd.DataFrame:
    valor_inicial = _get_valores_iniciais(df)
    cotas_normalizadas = df/valor_inicial
    return cotas_normalizadas

def calcula_retorno_janelas_moveis(cotas_diarias: pd.DataFrame, HP: int, dados_diarios_benchmark: pd.DataFrame) -> pd.DataFrame:
    '''Função que calcula o retorno em janelas móveis de periodo HP (holding period) de um ou mais fundos e seu benchmark.
    Parâmetros:
    -cotas_diarias (dataframe): com valor das cotas;
    -HP (int): periodo da janela móvel, em dias;
    -dados_diarios_benchmark (dataframe): índice diário do benchmark
    Retorno:
    -retorno_janelas_moveis (dataframe): retorno na janela móvel HP'''
    dados = pd.concat([cotas_diarias, dados_diarios_benchmark], axis=1).dropna()
    cotas_normalizadas = _get_cotas_normalizadas(dados)
    retorno_janelas_moveis = cotas_normalizadas.dropna().pct_change(HP)
    return retorno_janelas_moveis.sort_index().dropna()
   
def calcula_risco_retorno_fundos(
                                dados_fundos_cvm: pd.DataFrame,
                                ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    '''Função que extrai diversas informações de retorno e risco para uma primeira análise dos fundos.
    Parâmetros:
    -dados_fundos_cvm (dataframe): série temporal contendo as cotas diarias dos fundos. Cada coluna deve ser a cota de um fundo e índice as datas.
    Retorno:
    -risco_retorno (dataframe): contém volatilidade anualizada e retorno de todo o periodo também anualizado;
    -rentabilidade_fundos_diaria (dataframe): retorno diário dos fundos;
    -cotas_normalizadas (dataframe): valor das cotas diárias normalizada começando com valor unitário;
    -rentabilidade_fundos_acumulada (dataframe): rentabilidade acumulada de todo o período;
    -rentabilidade_acumulada_por_ano (dataframe): rentabilidade acumulada por cada ano;
    -rentabilidade_fundos_total (dataframe): rentabilidade total de todo o periodo'''
    serie_fundos = dados_fundos_cvm.copy()
    cotas_normalizadas = _get_cotas_normalizadas(serie_fundos)
    rentabilidade_fundos_diaria = cotas_normalizadas.pct_change()
    rentabilidade_fundos_acumulada = (1 + rentabilidade_fundos_diaria).cumprod() - 1

    T = serie_fundos.shape[0]
    valores_finais_nao_nulos = pd.DataFrame(_get_valores_finais(cotas_normalizadas), index=cotas_normalizadas.columns).T
    retorno_periodo_anualizado = (((valores_finais_nao_nulos/ _get_valores_iniciais(cotas_normalizadas))** (252 / T)- 1)).T
    retorno_periodo_anualizado.columns = ["rentabilidade"]

    rentabilidade_acumulada_por_ano = rentabilidade_fundos_acumulada.groupby(pd.Grouper(freq="Y")).last(1).T
    rentabilidade_acumulada_por_ano.columns = [str(x)[:4] for x in rentabilidade_acumulada_por_ano.columns]

    volatilidade_fundos = rentabilidade_fundos_diaria.std().to_frame() * np.sqrt(252)
    volatilidade_fundos.columns = ["volatilidade"]

    risco_retorno = pd.concat([volatilidade_fundos, retorno_periodo_anualizado], axis=1)
    return (
            risco_retorno.dropna().sort_values("rentabilidade", ascending=False),
            rentabilidade_fundos_diaria,
            cotas_normalizadas,
            rentabilidade_fundos_acumulada,
            rentabilidade_acumulada_por_ano
            )

def remove_outliers(df: pd.DataFrame, q: float = 0.05) -> pd.DataFrame:
    '''Remove outliers ao informar usando método do range interquartil, ou seja, retira os dados 
    acima de 1.5*Q3 (1-q) e abaixo de 1.5*Q1 (q), onde q é o quantil.
    Quanto maior o valor de q, mais dados serão removidos.'''
    df1 = df.copy()
    col = df1.columns.tolist()
    upper = df1[col].quantile(1-q)
    lower = df1[col].quantile(q)
    df2 =  df1[(df1[col] < upper) & (df1[col] > lower)]
    return df2

def plotar_comparacao_risco_retorno(
                                    df: pd.DataFrame,
                                    risco_retorno_carteira: Union[Tuple[int, int], None] = None,
                                    risco_retorno_benchmark: Union[Tuple[int, int], None] = None,
                                    nome_carteira: Union[str, None] = None,
                                    nome_benchmark: Union[str, None] = None,
                                    **opcionais: Any
                                    ) -> None:
    if risco_retorno_carteira:
        risco_carteira, retorno_carteira = risco_retorno_carteira
    if risco_retorno_benchmark:
        risco_benchmark, retorno_benchmark = risco_retorno_benchmark

    plt.figure(figsize=(opcionais.get("figsize")))
    sns.scatterplot(
        data=df, y="rentabilidade", x="volatilidade", alpha=0.45, color="gray"
    )
    if risco_retorno_carteira:
        plt.scatter(x=risco_carteira, y=retorno_carteira, marker="o", color="blue", s=200)
    if risco_retorno_benchmark:
        plt.scatter(x=risco_benchmark, y=retorno_benchmark, marker="o", color="red", s=200)
    plt.ylabel("Retorno (%aa)\n", rotation=0, labelpad=-70, loc="top")
    plt.xlabel("Volatilidade (%aa)")
    plt.suptitle('Risco x Retorno')
    plt.box(False)
    plt.grid(axis="y")
    plt.ylim(opcionais.get("ylim"))
    plt.xlim(opcionais.get("xlim"))
    if risco_retorno_carteira:
        plt.annotate(
            nome_carteira,
            xy=(risco_carteira + 0.5, retorno_carteira),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_carteira"),
            textcoords="offset points",
            color="darkblue",
            weight="bold",
            arrowprops=dict(arrowstyle="->", color="blue", connectionstyle="arc3,rad=-0.1"),
        )
    if risco_retorno_benchmark:
        plt.annotate(
            nome_benchmark,
            xy=(risco_benchmark + 0.5, retorno_benchmark),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_benchmark"),
            textcoords="offset points",
            color="darkred",
            weight="bold",
            arrowprops=dict(arrowstyle="->", color="r", connectionstyle="arc3,rad=-0.1"),
        )


def plotar_evolucao(
                    df: pd.DataFrame, lista_fundos: List[str], **opcionais: Any
                    ) -> Union[pd.DataFrame, None]:

    lista_fundos = [x.upper() for x in lista_fundos]
    cnpj = [x for x in df.columns.tolist() if x.split(" // ")[0] in lista_fundos]
    nome = [x for x in df.columns.tolist() if x.split(" // ")[1] in lista_fundos]
    if len(nome) == 0:
        for fundos in df.columns.tolist():
            fd = fundos.split(" // ")[1]
            for fds in lista_fundos:
                if fds in fd:
                    nome.append(fundos)
    colunas = nome + cnpj
    if len(colunas) > 2:
        data = df[colunas].dropna(axis=1, how="all")
        data = data[data > 0].dropna()
        maximo, minimo = (
            data.iloc[-1:].idxmax(axis=1).values[0],
            data.iloc[-1:].idxmin(axis=1).values[0],
        )
        legenda_maximo = maximo.split(" // ")[0]
        legenda_minimo = minimo.split(" // ")[0]
        data.index = pd.to_datetime(data.index)
        plt.figure(figsize=(opcionais.get("figsize")))
        plt.plot(
            data.drop([maximo, minimo], axis=1),
            color=opcionais.get("color"),
            alpha=opcionais.get("alpha"),
        )
        plt.plot(data[maximo], color=opcionais.get("color_maximo"))
        plt.plot(data[minimo], color=opcionais.get("color_minimo"))
        plt.ylabel("Cotas\n", rotation=0, labelpad=-20, loc="top")
        plt.xlabel("")
        plt.box(False)
        plt.grid(axis="y")
        plt.ylim(opcionais.get("ylim"))
        plt.xlim(opcionais.get("xlim"))
        plt.annotate(
            legenda_maximo,
            xy=(data.index[-1], data.iloc[-1:].max(axis=1).values[0]),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_maximo"),
            textcoords="offset points",
            color=opcionais.get("color_maximo"),
            weight="bold",
            arrowprops=dict(
                arrowstyle="->",
                color=opcionais.get("color_seta_maximo"),
                connectionstyle="arc3,rad=-0.1",
            ),
        )
        plt.annotate(
            legenda_minimo,
            xy=(data.index[-1], data.iloc[-1:].min(axis=1).values[0]),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_minimo"),
            textcoords="offset points",
            color=opcionais.get("color_minimo"),
            weight="bold",
            arrowprops=dict(
                arrowstyle="->",
                color=opcionais.get("color_seta_minimo"),
                connectionstyle="arc3,rad=-0.1",
            ),
        )
        return data
    elif len(colunas) <=2:
        data = df[colunas].dropna(axis=1, how="all")
        data = data[data > 0].dropna()
        data.index = pd.to_datetime(data.index)
        plt.figure(figsize=(opcionais.get("figsize")))
        plt.plot(
            data,
            color=opcionais.get("color"),
            alpha=opcionais.get("alpha"),
        )
        plt.ylabel("Cotas\n", rotation=0, labelpad=-20, loc="top")
        plt.xlabel("")
        plt.box(False)
        plt.grid(axis="y")
        plt.ylim(opcionais.get("ylim"))
        plt.xlim(opcionais.get("xlim"))
        return data
    elif not colunas:
        print("Fundo não encontrado")
        return None

def plotar_rentabilidade_janela_movel(df: pd.DataFrame, HP: int, benchmarks: pd.DataFrame) -> None:
    for fundo in df.columns:
        retorno = calcula_retorno_janelas_moveis(df[[fundo]], HP, benchmarks)

        plt.figure(figsize=(15, 6))
        plt.suptitle(f"Retorno de {HP} dias")
        plt.plot(retorno[fundo].multiply(100), lw=2)
        plt.plot(retorno[benchmarks.columns.tolist()].multiply(100), lw=1, alpha=0.8)
        plt.xticks(rotation=45)
        plt.xlabel('')
        plt.ylabel("%\n", rotation=0, labelpad=-15, loc="top")
        plt.legend([fundo.split("//")[-1]]+benchmarks.columns.tolist(), loc="upper right", frameon=False,
                     bbox_to_anchor=(0.5, 0.67, 0.5, 0.5))
        plt.tight_layout()
        plt.box(False)
        plt.grid(axis="x")
    plt.show()

def plotar_comparacao_risco_retorno(
                                    df: pd.DataFrame,
                                    risco_retorno_carteira: Union[Tuple[int, int], None] = None,
                                    risco_retorno_benchmark: Union[Tuple[int, int], None] = None,
                                    nome_carteira: Union[str, None] = None,
                                    nome_benchmark: Union[str, None] = None,
                                    **opcionais: Any
                                    ) -> None:
    if risco_retorno_carteira:
        risco_carteira, retorno_carteira = risco_retorno_carteira
    if risco_retorno_benchmark:
        risco_benchmark, retorno_benchmark = risco_retorno_benchmark

    plt.figure(figsize=(opcionais.get("figsize")))
    sns.scatterplot(
        data=df, y="rentabilidade", x="volatilidade", alpha=0.45, color="gray"
    )
    if risco_retorno_carteira:
        plt.scatter(x=risco_carteira, y=retorno_carteira, marker="o", color="blue", s=200)
    if risco_retorno_benchmark:
        plt.scatter(x=risco_benchmark, y=retorno_benchmark, marker="o", color="red", s=200)
    plt.ylabel("Retorno (%aa)\n", rotation=0, labelpad=-70, loc="top")
    plt.xlabel("Volatilidade (%aa)")
    plt.box(False)
    plt.grid(axis="y")
    plt.ylim(opcionais.get("ylim"))
    plt.xlim(opcionais.get("xlim"))
    if risco_retorno_carteira:
        plt.annotate(
            nome_carteira,
            xy=(risco_carteira + 0.5, retorno_carteira),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_carteira"),
            textcoords="offset points",
            color="darkblue",
            weight="bold",
            arrowprops=dict(arrowstyle="->", color="blue", connectionstyle="arc3,rad=-0.1"),
        )
    if risco_retorno_benchmark:
        plt.annotate(
            nome_benchmark,
            xy=(risco_benchmark + 0.5, retorno_benchmark),
            xycoords="data",
            xytext=opcionais.get("posicao_texto_benchmark"),
            textcoords="offset points",
            color="darkred",
            weight="bold",
            arrowprops=dict(arrowstyle="->", color="r", connectionstyle="arc3,rad=-0.1"),
        )

def supera_benchmark(dados: pd.DataFrame, benchmarks: pd.DataFrame, HP: int,
                     limit: float = 0.6, janela_analise: int = 252) -> pd.DataFrame:
    '''Função que indica a porcentagem de vezes em que o fundo supera o benchmark no periodo selecionado.
    O corte é dado pelo parâmetro limit.'''
    percentuais = pd.DataFrame()
    lista_benchmarks = [x for x in benchmarks.columns if 'Retorno' not in x]
    for i, fundo in enumerate(tqdm(dados.columns.tolist())):
        df1 = pd.DataFrame()
        for bench in lista_benchmarks:
            retorno = calcula_retorno_janelas_moveis(dados[[fundo]], HP, benchmarks[[bench]])
            retorno = retorno.sort_index().dropna()
            if retorno.empty: break
            elif retorno.shape[0] < janela_analise: continue
            eventos = (retorno[fundo] - retorno[bench])
            percentual_sobre_bench = len(eventos[eventos>0])/len(eventos)
            if percentual_sobre_bench >= limit:
                df = pd.DataFrame([fundo, percentual_sobre_bench*100], index=["Fundo", f"Supera {bench} (%)"]).T.set_index("Fundo")
                df1 = pd.concat([df1, df], axis=1)
        percentuais = pd.concat([percentuais, df1], axis=0)
    return percentuais[(percentuais>=limit*100)].dropna().sort_values(percentuais.columns.tolist(), ascending=False)

def qto_supera_benchmark(dados: pd.DataFrame, benchmarks: pd.DataFrame, HP: int, corte_bench: float=100, bench_corte='CDI') -> pd.DataFrame:
    '''Enquanto a função *supera_benchmark* indica se um fundo supera o benchmark.
    Essa função exibe quantas vezes os fundos superaram o benchmark, em média, e o quanto eles 
    ficam abaixo do benchmark, em média.
    Os fundos são ranqueados por % do benchmark. O parâmetro corte_bench filtra os fundos que performam,
    pelo menos, o mesmo que o fundo.'''
    percentuais = pd.DataFrame()
    lista_benchmarks = [x for x in benchmarks.columns if 'Retorno' not in x]
    for i, fundo in enumerate(tqdm(dados.columns.tolist())):
        df1 = pd.DataFrame()
        for bench in lista_benchmarks:
            retorno = calcula_retorno_janelas_moveis(dados[[fundo]], HP, benchmarks[[bench]])
            retorno = retorno.sort_index().dropna()
            if retorno.empty: break
            eventos = (retorno[fundo] - retorno[bench])
            media_sup = eventos[eventos>=0].mean()
            media_inf = eventos[eventos<0].mean()
            em_rel_cdi = (retorno[fundo]/retorno[bench]).mean()
            df = pd.DataFrame([fundo, media_sup, media_inf, em_rel_cdi], index=["Fundo", f"% de vezes, em média, acima {bench} (%)",
                                                                        f"% de vezes, em média, abaixo {bench} (%)",
                                                                        f'% do {bench}, em média']).T.set_index("Fundo")
            df1 = pd.concat([df1, df], axis=1)
        percentuais = pd.concat([percentuais, df1], axis=0)
    cols1 = [x for x in percentuais.columns if 'acima' not in x and 'abaixo' not in x]
    percentuais = (percentuais.sort_values(cols1, ascending=False)*100).fillna(0)
    if not bench_corte:
        for bench in lista_benchmarks:
            percentuais = percentuais[percentuais[f'% do {bench}, em média']>=corte_bench]
    else:
        percentuais = percentuais[percentuais[f'% do {bench_corte}, em média']>=corte_bench]
    return percentuais

def _repetir_elemento(seq):
    return cycle(seq)

def _traduz_frequencia(frequencia: str) -> List[str]:
    if 'month' in frequencia.lower().split(' ')[0]:
        freq = ["Meses", "Mensais"]
    elif 'year' in frequencia.lower().split(' ')[0]:
        freq = ["Anos", "Anuais"]
    elif 'sem' in frequencia.lower().split(' ')[0]:
        freq = ["Semestres", "Semestrais"]
    elif 'quarter' in frequencia.lower().split(' ')[0]:
        freq = ["Trimestres", "Trimestrais"]
    else:
        freq = [frequencia]
    return freq

def calcula_rentabilidade_periodo(dados_diarios: pd.DataFrame, freq: str = "M") -> pd.DataFrame:
    if freq.lower()=="sem":
        freq = "Q"
        inicio = dados_diarios.resample(f"{freq}S").first()
        fim = dados_diarios.resample(f"{freq}E").last()
        inicio = inicio[inicio.index.month.isin([1,7])]
        fim = fim[fim.index.month.isin([6,12])]
    else:
        inicio = dados_diarios.resample(f"{freq}S").first()
        fim = dados_diarios.resample(f"{freq}E").last()
    rentabilidade_periodo_total = pd.DataFrame()
    for (init, end) in zip(inicio.index, fim.index):
        df = dados_diarios[(dados_diarios.index>=init) & (dados_diarios.index<=end)].ffill().pct_change()
        rentabilidade_periodo = ((1 + df).cumprod() - 1).tail(1)
        rentabilidade_periodo.index = [end]
        rentabilidade_periodo_total = pd.concat([rentabilidade_periodo_total, rentabilidade_periodo])
    return rentabilidade_periodo_total.asfreq(f"{freq}E")

def _retorno_heatmap(dados_diarios: pd.DataFrame, period: str, nome: str) -> Union[pd.DataFrame, List[str]]:
    returns = calcula_rentabilidade_periodo(dados_diarios, period.upper())*100
    frequencia = period if period=='sem' else str(returns.index.freq).replace('<','').replace('>','')
    freq = _traduz_frequencia(frequencia)
    f = _repetir_elemento(list(range(1, returns.index.month[:-1].nunique()+1)))
    lista_repeticao = [next(f) for x in range(returns.shape[0])]
    
    returns["Ano"] = returns.index.year
    returns[f"{freq[0]}"] = lista_repeticao
    returns = returns.pivot_table(index=f"{freq[0]}", columns="Ano", values=nome, aggfunc="last").fillna(0).T
    return returns, freq

def plotar_heatmap_rentabilidade(dados_diarios: pd.DataFrame, period: str = 'M') -> None:
    '''Função que plota gráfico tipo heatmap que exibe o desempenho do fundo ou do benchmark no periodo indicado, que pode ser:
    -mensal (M), 
    -trimestral (Q), 
    -semestral (sem) ou 
    -anual (Y)
    Como entrada de dados, a função precisa do dataframe do retorno diário do fundo ou do benchmark.
    '''
    cmap = LinearSegmentedColormap.from_list(name='t',
        colors=["red", "white", 'green']
    )
    nome = dados_diarios.columns[0]
    returns, freq = _retorno_heatmap(dados_diarios, period, nome)

    grayscale=False
    ylabel=True
    fontname="Arial"
    annot_size=10

    fig, ax = plt.subplots(figsize=(10,5))
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)

    fig.set_facecolor("white")
    ax.set_facecolor("white")

    ax.set_title(
         f"Retornos {freq[1]} (%) - {nome.title()}\n",
        fontsize=14,
        y=0.995,
        fontname=fontname,
        fontweight="bold",
        color="black",
    )
    ax = sns.heatmap(
                returns,
                ax=ax,
                annot=True,
                center=0,
                annot_kws={"size": 10},
                fmt="0.2f",
                linewidths=0.5,
                square=False,
                cbar=True,
                cmap=cmap,
                cbar_kws={"format": "%.0f%%"},
            )

    # align plot to match other
    if ylabel:
        ax.set_ylabel("Ano", fontname=fontname, fontweight="bold", fontsize=12, rotation=0)
        ax.yaxis.set_label_coords(-0.1, 0.5)

    ax.tick_params(colors="#808080")
    plt.xticks(rotation=0, fontsize=annot_size * 1.2)
    plt.yticks(rotation=0, fontsize=annot_size * 1.2)

    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass
    try:
        fig.tight_layout(w_pad=0, h_pad=0)
    except Exception:
        pass

    plt.show(block=False)
    plt.close()

def plotar_heatmap_comparar_benchmark(dados_diarios_fundos: pd.DataFrame,
                                      dados_diarios_benchmarks: pd.DataFrame,
                                      period: str = "M") -> None:
    '''Função que exibe gráfico heatmap para facilitar a comparação de desempenho de um fundo com seu benchmark.
    O gráfico uma coluna Ultrapassa Benchmark onde indica, em %, quanto o fundo superou o benchmark no periodo (period), que pode ser:
    -mensal (M), 
    -trimestral (Q), 
    -semestral (sem) ou 
    -anual (Y)
    superou = 100 * num_vezes_que_superou_no_periodo/total_de_periodos
    As cores do gráfico auxiliam na indicação dos periodos em que houve superação.
    '''
    bench = dados_diarios_benchmarks.columns[0]
    nome = dados_diarios_fundos.columns[0]
    returns1, freq = _retorno_heatmap(dados_diarios_fundos, period, nome)
    returns2, freq2 = _retorno_heatmap(dados_diarios_benchmarks, period, bench)
    
    df4 = returns1 - returns2
    df4['nao_superou'] = (df4 < 0).sum(axis=1)
    df4['superou'] = (df4[[df4.columns[0]]] > 0).sum(axis=1)
    df4[f'Ultrapassa {bench}'] = np.where(df4['superou']>0, 100*(df4['superou']/(df4['superou']+df4['nao_superou'])), 0)
    
    df = pd.concat([returns1, df4[f'Ultrapassa {bench}']], axis=1)
    colors = ["darkred", "red", 'lightcoral', "lightpink", "pink",  "lightgreen", 'limegreen','forestgreen', "darkgreen"]
    cmap_gradient = LinearSegmentedColormap.from_list("cmap", colors)

    last_column = df.iloc[:, -1]
    norm = plt.Normalize(last_column.min(), last_column.max())
    row_colors = cmap_gradient(norm(last_column))

    ylabel = True
    fontname = "Arial"
    annot_size = 10
    
    # Criar o gráfico
    fig, ax = plt.subplots(figsize=(12, 6))
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)
    
    fig.set_facecolor("white")
    ax.set_facecolor("white")

    ax.set_title(
        f"Retornos {freq[1]} (%) - {nome.title()}\n",
        fontsize=14,
        y=1.04,
        fontname=fontname,
        fontweight="bold",
        color="black",
    )

    heatmap = sns.heatmap(
        df,
        ax=ax,
        annot=False,  # Anotar os valores nas células
        fmt=".2f",  # Formato dos números anotados
        cmap=cmap_gradient,  # Usar o colormap personalizado
        cbar=True,  # Adicionar barra de cores lateral
        linewidths=1.5,  # Espaçamento entre as células
        linecolor="gray",  # Cor das bordas das células
        cbar_kws={"format": "%.0f%%"}, #"ticks": [-100, -50, 0, 50, 100]},
        square=False
    )

    def get_text_color(cell_color):
        """Retorna 'white' se a cor de fundo for escura, caso contrário 'black'."""
        # Converte a cor para HSV e verifica a luminância (V)
        hsv = rgb_to_hsv(np.array(cell_color)[:3])
        luminance = hsv[2]  # Valor V em HSV
        return "white" if luminance < 0.5 else "black"
    
    # Ajustar as anotações para incluir '%'
    for i in range(df.shape[0]):
        for j in range(df.shape[1]):
            value = df.iloc[i, j]
            text = f"{value:.2f}%"  # Adicionar o símbolo '%'
            cell_color = cmap_gradient(norm(df.iloc[i, -1]))
            text_color = get_text_color(cell_color)
            heatmap.text(
                j + 0.58,  # Posição x (coluna atual)
                i + 0.5,  # Posição y (linha atual)
                text,  # Valor formatado
                ha="center",  # Alinhamento horizontal
                va="center",  # Alinhamento vertical
                color=text_color,
                #color="black" if j == df.shape[1] - 1 else "darkslategray",  # Cor do texto
                fontsize=annot_size,  # Tamanho da fonte
                weight="bold" if j == df.shape[1] - 1 else "normal",  # Negrito na última coluna
            )

    for i, color in enumerate(row_colors):
        plt.gca().add_patch(plt.Rectangle((0, i), df.shape[1], 1, fill=True, color=color, lw=0))
    
    for i in range(df.shape[0] + 1):
        plt.axhline(i, color="white", linewidth=.5, linestyle="-")
    for j in range(df.shape[1] + 1):
        plt.axvline(j, color="white", linewidth=.5, linestyle="-")

    ax.set_xticks(np.arange(df.shape[1]-1)+0.45)
    ax.set_xticklabels(df.columns[:-1], rotation=0, fontsize=annot_size * 1.2)
    ax.tick_params(axis="x", which="major", pad=5)  # Aumentar o espaçamento da última coluna

    ax.text(
        df.shape[1] - 0.5,  # Posição x (última coluna)
        -.6,  # Posição y (acima do heatmap)
        df.columns[-1],  # Texto do rótulo
        ha="center",  # Alinhamento horizontal
        va="center",  # Alinhamento vertical
        fontsize=11,  # Tamanho da fonte
        fontweight="bold",  # Negrito
        color="black",  # Cor do texto
    )

    if ylabel:
        ax.set_ylabel("Ano", fontname=fontname, fontweight="bold", fontsize=12, rotation=0)
        ax.yaxis.set_label_coords(-0.1, 0.5)

    ax.tick_params(colors="#808080")
    plt.xticks(rotation=0, fontsize=annot_size * 1.2)
    plt.yticks(rotation=0, fontsize=annot_size * 1.2)
    
    try:
        plt.subplots_adjust(hspace=0, bottom=0, top=1)
    except Exception:
        pass
    try:
        fig.tight_layout(w_pad=0, h_pad=0)
    except Exception:
        pass
    
    plt.show(block=False)
    plt.close()


# -*- coding: utf-8 -*-
from .fundosbr import *
from .benchmarks import *
from .comparador import *
from . import version

__version__ = version.__version__
__author__ = "Rafael Rodrigues, rafa-rod @ GitHub"


# -*- coding: utf-8 -*-
"""
@author: Rafael
"""
import io
import time
import warnings
import zipfile
from datetime import datetime
from typing import List, Union, Dict, Optional
import polars as pl
import pandas as pd
import requests

warnings.filterwarnings("ignore")

pd.set_option("display.float_format", lambda x: "%.6f" % x)
pd.set_option("display.max_rows", 100)
pd.set_option("display.max_columns", 10)
pd.set_option("display.width", 1000)

def get_classes() -> List[str]:
    '''Lista as classes disponíveis para filtro.'''
    return ['Renda Fixa', 'Ações', 'Multimercado', 'Cambial', 'Curto Prazo', 'Referenciado']

def pontua_cnpj(cnpj: str) -> str:
    '''Efetua a pontuação do CNPJ'''
    if len(cnpj) < 14:
        cnpj = cnpj.zfill(14)
    cnpj = cnpj.replace("-","").replace(".","").replace("/","")
    p1, p2, p3, p4, p5 = cnpj[:2], cnpj[2:5], cnpj[5:8], cnpj[8:12], cnpj[12:]
    return f"{p1}.{p2}.{p3}/{p4}-{p5}"

def _get_response(url: str, proxy: Optional[Dict[str, str]] = None):
    if proxy:
        resposta = requests.get(url, proxies=proxy, verify=False)
    else:
        resposta = requests.get(url)
    return resposta

def _ler_zip_files(resposta, arquivo: str) -> pl.dataframe.frame.DataFrame:
    if str(resposta) != "<Response [200]>" and str(resposta) != '<Response [407]>':
        raise ValueError("Não foi possível baixar os dados solicitados")
    elif str(resposta) == '<Response [407]>':
        raise ValueError("Necessário informar proxy correta. Response [407]")
    zf = zipfile.ZipFile(io.BytesIO(resposta.content))
    zf = zf.open(arquivo)
    lines = zf.readlines()
    lines = [i.strip().decode("ISO-8859-1").split(";") for i in lines]
    fundos = pl.DataFrame(lines[1:], schema=lines[0])
    return fundos

def get_cadastro_fundos(
    classe: Optional[Union[List[str], str]] = None, 
    proxy: Optional[Dict[str, str]] = None,
    output_format: str = 'pandas') -> Union[pd.DataFrame, pl.dataframe.frame.DataFrame]:
    '''Busca o cadastro dos fundos em funcionamento normal, dos tipos de classe FIF e FIDC cuja classificação seja não nula
    e busca sua respectiva classe.'''
    classes_disponiveis = get_classes()
    start = time.time()
    url1 = "http://dados.cvm.gov.br/dados/FI/CAD/DADOS/cad_fi_hist.zip"
    resposta1 = _get_response(url1, proxy=proxy)
    url2 = "http://dados.cvm.gov.br/dados/FI/CAD/DADOS/registro_fundo_classe.zip"
    resposta2 = _get_response(url2, proxy=proxy)
    resposta3 = _get_response(url2, proxy=proxy)

    arquivo1, arquivo2, arquivo3 = "cad_fi_hist_classe.csv", "registro_classe.csv", "registro_fundo.csv"
    classes_dos_fundos = _ler_zip_files(resposta1, arquivo1)
    classes_dos_fundos = classes_dos_fundos.filter(pl.col('DT_FIM_CLASSE')!='') #classes atuais

    nome_dos_fundos = _ler_zip_files(resposta2, arquivo2)
    nome_dos_fundos = nome_dos_fundos.with_columns(pl.col(["CNPJ_Classe"]).map_elements(pontua_cnpj))

    fundos_filtrado = classes_dos_fundos.join(nome_dos_fundos, right_on='CNPJ_Classe', left_on='CNPJ_FUNDO', how='outer')
    fundos_filtrado = fundos_filtrado.with_columns(
                                        pl.when(pl.col('CNPJ_FUNDO').is_null())
                                          .then(pl.col('CNPJ_Classe'))
                                          .otherwise(pl.col('CNPJ_FUNDO'))
                                          .alias('CNPJ')
                                            )
    fundos_filtrado = fundos_filtrado.drop('CNPJ_FUNDO').rename({'CNPJ': 'CNPJ_FUNDO'})
    mais_info_dos_fundos = _ler_zip_files(resposta3, arquivo3)
    mais_info_dos_fundos = mais_info_dos_fundos.select(['CNPJ_Fundo', 'Tipo_Fundo', 'Denominacao_Social', 'Situacao', 'Data_Adaptacao_RCVM175'])
    mais_info_dos_fundos = mais_info_dos_fundos.rename({'CNPJ_Fundo': 'CNPJ_FUNDO'})
    mais_info_dos_fundos = mais_info_dos_fundos.with_columns(pl.col(["CNPJ_FUNDO"]).map_elements(pontua_cnpj))
    fundos_filtrado = fundos_filtrado.join(mais_info_dos_fundos, on=['CNPJ_FUNDO', 'Denominacao_Social', 'Situacao'], how='right')
    fundos_filtrado = fundos_filtrado.filter( (pl.col('Situacao')=="Em Funcionamento Normal") &
                                              (pl.col('Tipo_Fundo').is_in(['FIDC', 'FI', 'FIF']) )).drop(['CNPJ_Classe',
                                                                                                          'DT_REG',
                                                                                                        'DT_INI_CLASSE',
                                                                                                        'DT_FIM_CLASSE',
                                                                                                        'ID_Registro_Fundo',
                                                                                                        'ID_Registro_Classe'])
    if classe:
        if not isinstance(classe, list):
            classe = [classe]
        check_classes = [x for x in classe if x not in classes_disponiveis]
        if check_classes:
            raise ValueError(f"Classe não encontrada {check_classes}")
        fundos_filtrado = fundos_filtrado.filter((pl.col('CLASSE').is_in(classe)) | (pl.col('CLASSE').is_null()) |
                                                 (pl.col('CLASSE')==''))
    fundos_filtrado = fundos_filtrado.with_columns(pl.col('Denominacao_Social').str.to_uppercase())
    if output_format.lower() == 'pandas':
        fundos_filtrado = fundos_filtrado.to_pandas()
    print(f"Cadastro finalizado em {round((time.time()-start)/60,2)} minutos")
    return fundos_filtrado

def mesclar_bases(cadastro_fundos: pl.dataframe.frame.DataFrame, informe_diario_fundos: pl.dataframe.frame.DataFrame,
                  output_format: str = 'pandas') -> pl.dataframe.frame.DataFrame:
    '''Função para obter dados adicionais dos Fundos que estão em seu cadastro.
    Basta informar o dataframe do cadastro com o dataframe do informe diario para obter as informações.'''
    if isinstance(cadastro_fundos, pd.DataFrame):
        cadastro_fundos = pl.from_pandas(cadastro_fundos)
    if isinstance(informe_diario_fundos, pd.DataFrame):
        if 'DT_COMPTC' not in informe_diario_fundos.columns:
            informe_diario_fundos = informe_diario_fundos.reset_index()
        informe_diario_fundos = pl.from_pandas(informe_diario_fundos)
    dados_completos_filtrados = informe_diario_fundos.join(cadastro_fundos, on=['CNPJ_FUNDO'], how="inner")
    dados_completos_filtrados = dados_completos_filtrados.with_columns(((pl.col('CNPJ_FUNDO')) + ' // ' + (pl.col('Denominacao_Social'))).alias('CNPJ - Nome'))
    if output_format.lower() == 'pandas':
        return dados_completos_filtrados.to_pandas().set_index('DT_COMPTC').sort_index()
    else:
        return dados_completos_filtrados.sort('DT_COMPTC')

def _ler_dados_diarios(ano: int, mes: int, proxy: Optional[Dict[str, str]] = None,
                       cnpj: Optional[str] = None,
                       num_minimo_cotistas: Optional[int] = None, 
                       patriminio_liquido_minimo: Optional[int] = None) -> pl.dataframe.frame.DataFrame:
    arquivo = "inf_diario_fi_{:02d}{:02d}.csv".format(ano, mes) if ano> 2004 else "inf_diario_fi_{:02d}.csv".format(ano)
    url = "http://dados.cvm.gov.br/dados/FI/DOC/INF_DIARIO/DADOS/inf_diario_fi_{:02d}{:02d}.zip".format(ano, mes) if ano>= 2021 else \
           "http://dados.cvm.gov.br/dados/FI/DOC/INF_DIARIO/DADOS/HIST/inf_diario_fi_{:02d}.zip".format(ano)
    resposta = _get_response(url, proxy=proxy)
    fundos = _ler_zip_files(resposta, arquivo)
    cols1 = [x for x in fundos.columns if "TP_FUNDO" in x]
    if cols1 and int(ano)>=2004:
        fundos = fundos.rename({cols1[0]: 'TP_FUNDO'})
        fundos = fundos.filter(pl.col("TP_FUNDO").is_in(['FI','FIF','CLASSES - FIF']))
    cols2 = [x for x in fundos.columns if "CNPJ_FUNDO" in x][0]
    fundos = fundos.rename({cols2: 'CNPJ_FUNDO'})
    fundos = fundos.with_columns(pl.col("NR_COTST").cast(pl.Int32, strict=False))
    fundos = fundos.with_columns(pl.col("VL_PATRIM_LIQ").cast(pl.Float64, strict=False))
    fundos = fundos.with_columns(pl.col("VL_TOTAL").cast(pl.Float64, strict=False))
    fundos = fundos.with_columns(pl.col("CAPTC_DIA").cast(pl.Float64, strict=False))
    fundos = fundos.with_columns(pl.col("RESG_DIA").cast(pl.Float64, strict=False))
    fundos = fundos.with_columns(pl.col("VL_QUOTA").cast(pl.Float32, strict=False))
    fundos = fundos.with_columns(pl.col("DT_COMPTC").str.to_datetime("%Y-%m-%d"))
    if num_minimo_cotistas:
        fundos = fundos.filter(pl.col("NR_COTST") >= num_minimo_cotistas)
    if patriminio_liquido_minimo:
        fundos = fundos.filter(pl.col("VL_PATRIM_LIQ") >= patriminio_liquido_minimo)
    if cnpj:
        if isinstance(cnpj, str): cnpj = [cnpj]
        lista_cnpj = [pontua_cnpj(x) for x in cnpj]
        fundos = fundos.filter(pl.col("CNPJ_FUNDO").is_in(lista_cnpj))
    return fundos.select(['DT_COMPTC', 'CNPJ_FUNDO', 'NR_COTST', 'VL_PATRIM_LIQ', 'VL_QUOTA',
                          'VL_TOTAL', 'CAPTC_DIA', 'RESG_DIA']).unique().sort('DT_COMPTC')

def fundosbr(
            anos: Union[List[int], int],
            meses: Union[List[int], int],
            cnpj: Optional[str] = None,
            num_minimo_cotistas: Optional[int] = None,
            patriminio_liquido_minimo: Optional[int] = None,
            proxy: Optional[Dict[str, str]] = None,
            output_format: str = 'pandas'
				) -> Union[pd.DataFrame, pl.dataframe.frame.DataFrame]:
    start = time.time()
    if isinstance(anos, int): anos = [anos]
    else: anos = list(anos)
    if isinstance(meses, int): meses = [meses]
    else: anos = list(anos)
    informe_diario_fundos_historico = pl.DataFrame()
    for ano in anos:
        for mes in meses:
            if ano == datetime.now().year and mes <= datetime.now().month:
                informe_diario_fundos_filtrado = _ler_dados_diarios(ano, mes, proxy, cnpj, num_minimo_cotistas, patriminio_liquido_minimo)
                informe_diario_fundos_historico = pl.concat([informe_diario_fundos_historico, informe_diario_fundos_filtrado])
            elif ano < datetime.now().year:
                informe_diario_fundos_filtrado = _ler_dados_diarios(ano, mes, proxy, cnpj, num_minimo_cotistas, patriminio_liquido_minimo)
                informe_diario_fundos_historico = pl.concat([informe_diario_fundos_historico, informe_diario_fundos_filtrado])
    print(f"Dados diários finalizados em {round((time.time()-start)/60,2)} minutos")
    if output_format.lower() == 'pandas':
        return informe_diario_fundos_historico.to_pandas().set_index('DT_COMPTC').sort_index()
    else:
        return informe_diario_fundos_historico.sort('DT_COMPTC')

def get_fip(ano: int, 
            proxy: Union[Dict[str, str], None] = None) -> pd.DataFrame:
    start = time.time()
    url = "http://dados.cvm.gov.br/dados/FIP/DOC/INF_TRIMESTRAL/DADOS/inf_trimestral_fip_{:02d}.csv".format(ano)
    if proxy:
        try:
            dados_cvm = requests.get(url, proxies=proxy, verify=False)
        except AttributeError:
            raise ValueError("Necessário informar proxy correta.")
    else:
        dados_cvm = requests.get(url)
    if str(dados_cvm) == '<Response [404]>':
        raise ValueError("Não há dados para esta data. Response [404]")
    lines = [i.strip().split(";") for i in dados_cvm.text.split("\n")]
    end = time.time()
    print(f"Finalizado em {round((end-start)/60,2)} minutos")
    return pd.DataFrame(lines[1:], columns=lines[0])

def get_fidc(ano: int, 
             mes: int,
             tabela: str = 'X',
             subtabela: int = 3,
             proxy: Union[Dict[str, str], None] = None) -> pd.DataFrame:
    '''Função que busca dados de Fundos Estruturados - FIDC do FUNDOS.NET da CVM 
    que se enquadram na instrução 175 da CVM, exemplo:
    https://fnet.bmfbovespa.com.br/fnet/publico/abrirGerenciadorDocumentosCVM?cnpjFundo=11701985000107
    Outros fundos de crédito, até mesmo com FIDC no nome, podem estar nos informes diários.
    A tabela I exibe dados gerais do fundo, enquanto a tabela X subtabela 3 indica a rentabilidade mensal.
    Para demais tabelas, consultar site da CVM.
    '''
    start = time.time()
    url = "http://dados.cvm.gov.br/dados/FIDC/DOC/INF_MENSAL/DADOS/inf_mensal_fidc_{:02d}{:02d}.zip".format(ano, mes)
    if proxy:
        try:
            dados_cvm = requests.get(url, proxies=proxy, verify=False)
        except AttributeError:
            raise ValueError("Necessário informar proxy correta.")
    else:
        dados_cvm = requests.get(url)
    if str(dados_cvm) == '<Response [404]>':
        raise ValueError("Não há dados para esta data. Response [404]")
    if tabela.upper()=='X':
        arquivo = f"inf_mensal_fidc_tab_{tabela}_{subtabela}_{ano:02d}{mes:02d}.csv"
    else:
        arquivo = f"inf_mensal_fidc_tab_{tabela}_{ano:02d}{mes:02d}.csv"
    zf = zipfile.ZipFile(io.BytesIO(dados_cvm.content))
    zf = zf.open(arquivo)
    lines = zf.readlines()
    lines = [i.strip().decode("ISO-8859-1").split(";") for i in lines]
    end = time.time()
    print(f"Finalizado em {round((end-start)/60,2)} minutos")
    return pd.DataFrame(lines[1:], columns=lines[0])